/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifdef _USRDLL
  #define _QUBFITS_PY_
#endif
#include "_qubfits.h"

#include <algorithm>

void TreeToQFData(QUB_Tree tree, QF_Data* data, double** ff)
{
  data->n  = tree["n"].dataAsInt();
  data->xx = (double*) tree["Series"][ tree["XName"].dataAsString() ].data();
  data->yy = (double*) tree["Series"][ tree["YName"].dataAsString() ].data();
  data->ww = (double*) tree["ww"].data();
  if ( tree["ff"].dataCount() < data->n )
    tree["ff"].setNumData(QTR_TYPE_DOUBLE, 1, 1, (void*)0);
  *ff = (double*) tree["ff"].data();

  data->nv = 0;
  QUB_TreeMonoIter si = tree["Series"].children();
  for (data->nv = 0; ! si->isNull(); ++si)
    if ( si->dataCount() >= data->n )
      ++(data->nv);

  data->vname = new char*[data->nv];
  data->vvv   = new double*[data->nv];
  int i = 0;
  for (si = tree["Series"].children(); ! si->isNull(); ++si, ++i) {
    int len = si->name().size();
    data->vname[i] = new char[len+1];
    memcpy(data->vname[i], si->name().c_str(), len);
    data->vname[i][len] = 0;
    data->vvv[i] = (double*) si->data();
  }
}

void CleanupQFData(QF_Data* data)
{
  for (int i=0; i<data->nv; ++i) {
    delete [] data->vname[i];
  }
  delete [] data->vname;
  delete [] data->vvv;
}

// ******************************************************************************************************************

typedef struct {
  int Cls, N;
} NCls;

typedef struct {
  int First, Last;
  double Mean, Std;
} ResIval;

class ResTree {
public:
  int First, Last;
  double Mean, Std;
  ResTree *Left, *Right;
  
  ResTree(int aFirst, int aLast)
    : First(aFirst), Last(aLast), Left(NULL), Right(NULL)
  {}
  ~ResTree() {
    if ( Left )   delete Left;
    if ( Right )  delete Right;
  }
  
  void split() {
    // Assert((Left = nil) and (Right = nil), 'TAdResTree.Split: not a leaf node');
    int count = Last - First + 1;
    splitBefore(First + (count / 2));
  }
  void splitBefore(int RightFirst) {
    Left  = new ResTree(First, RightFirst-1);
    Right = new ResTree(RightFirst, Last);
  }
  void calcMeanStd(double* data) {
    int n = Last - First + 1;
    if ( n == 1 ) {
      Mean = data[First];
      Std  = 0.0;
      return;
    }
    
    Mean = 0.0;
    for (int i=First; i<=Last; ++i)
      Mean += data[i];
    Mean = Mean / n;
    
    Std = 0.0;
    for (int i=First; i<=Last; ++i) {
      double x = data[i] - Mean;
      Std += x*x;
    }
    Std = sqrt(Std / (n-1));
  }
  void splitStd(double* data, double MaxStd) {
    if ( ! Left ) {
      calcMeanStd(data);
      if ( Std > MaxStd )
	split();
    }
    if ( Left ) {
      Left->splitStd(data, MaxStd);
      Right->splitStd(data, MaxStd);
    }
  }
};

class ResTreeDFS {
private:
  bool LeavesOnly;
  vector<ResTree*> ToVisit;
  
  void push(ResTree* node) { ToVisit.push_back(node); }
  ResTree* pop() {
    ResTree* result = ToVisit.back();
    ToVisit.pop_back();
    return result;
  }

public:
  ResTree* Here;
  
  ResTreeDFS(ResTree* Root, bool aLeavesOnly)
    : LeavesOnly(aLeavesOnly), Here(NULL)
  {
    push(Root);
  }
  bool next() { // returns false if no more
    bool result = ToVisit.size();
    if ( result ) {
      Here = pop();
      if ( Here->Left ) {
	push(Here->Right);
	push(Here->Left);
	if ( LeavesOnly )
	  result = next();
      }
    }
    return result;
  }
};

//******************************************************************************

bool QF_AdaptiveResample_closest(double* data, int n, double MaxStd,
			 int& resN, vector<ResIval>& intervals, vector<int>& useSamples)
{
  ResTree *tree = new ResTree(0, n-1);
  tree->splitStd(data, MaxStd);
  
  int i=0;
  ResTreeDFS *trav = new ResTreeDFS(tree, true);
  while ( trav->next() ) {
    intervals.resize(i+1);
    ResIval& ival = intervals[i];
    ival.First = trav->Here->First;
    ival.Last  = trav->Here->Last;
    ival.Mean  = trav->Here->Mean;
    ival.Std   = trav->Here->Std;
    
    int min_ix = ival.First;
    double min_v = fabs(ival.Mean - data[ival.First]);
    for (int j=ival.First+1; j<=ival.Last; ++j) {
      double d = fabs(ival.Mean - data[j]);
      if ( d < min_v ) {
	min_ix = j;
	min_v  = d;
      }
    }
    useSamples.push_back(min_ix);

    ++i;
  }
  resN = i;
  delete trav;
  delete tree;
  return false;
}


// ******************************************************************************************************************

class QF_DataPt {
public:
  int i;
  double v;
  QF_DataPt(int ii, double vv)
    : i(ii), v(vv) {}
};
bool operator< (const QF_DataPt& a, const QF_DataPt& b) {
  return (a.v < b.v);
}

QUB_Tree QF_UseData(QUB_Tree treeData, QUB_Tree use)
{
  bool absX = treeData["AbsX"].dataAsInt();
  QF_Data origData;
  double *ff;
  TreeToQFData(treeData, &origData, &ff);
  
  // start with the raw data (copy)
  treeData["XName"].setData( use["XName"].dataAsString() );
  treeData["YName"].setData( use["YName"].dataAsString() );
  QUB_Tree result = treeData.clone();
  // pick out xx[], vv[][]  QF_Data data;
  int n = result["n"].dataAsInt();
  int selCount = n;
  double *xx = (double*) result["Series"][ result["XName"].dataAsString() ].data();
  vector<double*> vv;
  for (QUB_TreeMonoIter si = result["Series"].children(); ! si->isNull(); ++si)
    if ( si->dataCount() >= n )
      vv.push_back( (double*) si->data() );
  // sort by increasing x
  vector<QF_DataPt> sorts;
  for (int i=0; i<n; ++i)
    sorts.push_back(QF_DataPt(i, xx[i]));
  sort(sorts.begin(), sorts.end());
  for (int i=0; i<n; ++i) {
    xx[i] = origData.xx[ sorts[i].i ];
    for (int j=0; j<vv.size(); ++j)
      vv[j][i] = origData.vvv[j][ sorts[i].i ];
  }
  // calc ww[]
  if ( result["ww"].dataCount() < n )
    result["ww"].setNumData(QTR_TYPE_DOUBLE, n, 1, (void*) 0);
  double *ww = (double*) result["ww"].data();
  double *wws = (double*) result["Series"].find( use["WName"].dataAsString() )->data();
  if ( ! wws ) {
    for (int i=0; i<n; ++i)
      ww[i] = 1.0;
  }
  else if ( use["Winverse"].dataAsInt() ) {
    for (int i=0; i<n; ++i)
      if ( wws[i] != 0.0 )
	ww[i] = 1.0 / wws[i];
      else
	ww[i] = 0.0;
  }
  else {
    for (int i=0; i<n; ++i)
      ww[i] = wws[i];
  }
  // filter?
  // filter only uniformly-sampled data
  bool canFilter = n >= 10;
  double sampling = 0.0;
  double freq = use["Freq"].dataAsDouble();
  if ( canFilter ) {
    double lo = xx[1] - xx[0];
    double hi = lo;
    for (int i=2; i<n; ++i) {
      double x = xx[i] - xx[i-1];
      lo = min(x, lo);
      hi = max(x, hi);
    }
    canFilter = fabs(lo - hi) < 1.0e-8;
    sampling = lo;
    // min freq fits 4 periods in data
    freq = max(freq, 4.0/(xx[n-1] - xx[0]));
  }
  if ( canFilter && use["Filter"].dataAsInt() ) {
    vector<double> filtered(n);
    for (int i=0; i<vv.size(); ++i)
      if (vv[i] != xx)
	if (0 == filterdatad(vv[i], &(filtered[0]), n, sampling, freq))
	  // overwrite data with filtered on success
	  memcpy(vv[i], &(filtered[0]), n*sizeof(double));
    if (0 == filterdatad(ww, &(filtered[0]), n, sampling, freq))
      // this is probably not the right way to "filter" the weights
      memcpy(ww, &(filtered[0]), n*sizeof(double));
    
    // decimate by largest D: (2*Freq) < 1/(Sampling*D)
    //   Samp*D < 1/(2*Freq)
    //   D < 1/(2*Freq*Samp); D = floor(1/(2*Freq*Samp))
    int filterDec = max(1, int(floor(1.0/(2.0*freq*sampling))));
    selCount = n / filterDec;
    for (int i=0; i<selCount; ++i) {
      for (int j=0; j<vv.size(); ++j)
	vv[j][i] = vv[j][i*filterDec];
      ww[i] = ww[i*filterDec];
    }
    // take the last point if it's not there
    if ( ((selCount-1) * filterDec) < (n-1) ) {
      for (int j=0; j<vv.size(); ++j)
	vv[j][selCount] = vv[j][n-1];
      ww[selCount] = ww[n-1];
      ++selCount;
    }
  }

  // measure data
  int A = 0;
  double X = xx[0] + (use["SelFrom"].dataAsDouble() / 100.0) * (xx[n-1] - xx[0]);
  double start = absX ? 0.0 : X;
  while ( (A < selCount) && (xx[A] < X) )
    ++A;
  int B = selCount - 1;
  X = xx[0] + (use["SelTo"].dataAsDouble() / 100.0) * (xx[n-1] - xx[0]);
  while ( (B >= 0) && (xx[B] > X) )
    --B;
  selCount = B - A + 1;

  result["n"].setData(QTR_TYPE_INT, selCount);
  result["Xoff"].setData(QTR_TYPE_DOUBLE, start);

  // resample?
  bool resample = use["Resample"].dataAsInt();
  if ( resample ) {
    int resN;
    vector<ResIval> intervals;
    vector<int> useSamples;
    try {
      double *yy = (double*) result["Series"][ result["YName"].dataAsString() ].data();
      QF_AdaptiveResample_closest(yy, selCount, use["ResampleStd"].dataAsDouble(1.0),
			          resN, intervals, useSamples);
      for (int i=0; i<resN; ++i) {
	for (int j=0; j<vv.size(); ++j)
	  if (vv[j] != xx)
	    vv[j][i] = vv[j][A+useSamples[i]];
	xx[i] = xx[A+useSamples[i]] - start;
        // fill WW[i] with mean of WW[j] in interval
	double w = 0.0;
	for (int j=intervals[i].First; j<=intervals[i].Last; ++j)
	  w += ww[j];
	ww[i] = w / (intervals[i].Last - intervals[i].First + 1);
      }
      selCount = resN;
      result["n"].setData(QTR_TYPE_INT, resN);
    } catch (...) {
      resample = false;
    }
  }
  if ( ! resample ) {
    for (int j=0; j<vv.size(); ++j)
      if ( vv[j] != xx )
	memmove(vv[j], vv[j]+A, selCount*sizeof(double));
    memmove(ww, ww+A, selCount*sizeof(double));
    for (int i=0; i<selCount; ++i)
      xx[i] = xx[A+i] - start;
  }

  // cout << "using " << selCount << " data points" << endl;
  // now re-alloc fit curve
  result["ff"].setNumData(QTR_TYPE_DOUBLE, selCount, 1, (void*)0);

  CleanupQFData(&origData);
  return result;
}


//******************************************************************************
